@@ -1,7 +1,8 @@
 // qsamplerChannel.cpp
 //
 /****************************************************************************
-   Copyright (C) 2003-2005, rncbc aka Rui Nuno Capela. All rights reserved.
+   Copyright (C) 2004-2007, rncbc aka Rui Nuno Capela. All rights reserved.
+   Copyright (C) 2007, Christian Schoenebeck
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
@@ -13,36 +14,40 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+   You should have received a copy of the GNU General Public License along
+   with this program; if not, write to the Free Software Foundation, Inc.,
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
 *****************************************************************************/
 
 #include "qsamplerAbout.h"
 #include "qsamplerChannel.h"
+#include "qsamplerUtilities.h"
 
 #include "qsamplerMainForm.h"
 #include "qsamplerChannelForm.h"
 
-#include <qfileinfo.h>
-#include <qcombobox.h>
+#include <QFileInfo>
+#include <QComboBox>
 
 #ifdef CONFIG_LIBGIG
 #include "gig.h"
 #endif
 
+namespace QSampler {
+
 #define QSAMPLER_INSTRUMENT_MAX 100
 
+#define UNICODE_RIGHT_ARROW	QChar(char(0x92), char(0x21))
+
 
 //-------------------------------------------------------------------------
-// qsamplerChannel - Sampler channel structure.
+// QSampler::Channel - Sampler channel structure.
 //
 
 // Constructor.
-qsamplerChannel::qsamplerChannel ( qsamplerMainForm *pMainForm, int iChannelID )
+Channel::Channel ( int iChannelID )
 {
-	m_pMainForm  = pMainForm;
 	m_iChannelID = iChannelID;
 
 //  m_sEngineName       = noEngineName();
@@ -54,58 +59,36 @@ qsamplerChannel::qsamplerChannel ( qsamplerMainForm *pMainForm, int iChannelID )
 	m_iMidiDevice       = -1;
 	m_iMidiPort         = -1;
 	m_iMidiChannel      = -1;
+	m_iMidiMap          = -1;
 	m_sAudioDriver      = "ALSA";
 	m_iAudioDevice      = -1;
-	m_fVolume           = 0.0;
+	m_fVolume           = 0.0f;
 	m_bMute             = false;
 	m_bSolo             = false;
 }
 
 // Default destructor.
-qsamplerChannel::~qsamplerChannel (void)
+Channel::~Channel (void)
 {
 }
 
 
-// Main application form accessor.
-qsamplerMainForm *qsamplerChannel::mainForm(void) const
-{
-	return m_pMainForm;
-}
-
-
-// The global options settings delegated property.
-qsamplerOptions *qsamplerChannel::options (void) const
-{
-	if (m_pMainForm == NULL)
-		return NULL;
-
-	return m_pMainForm->options();
-}
-
-
-// The client descriptor delegated property.
-lscp_client_t *qsamplerChannel::client (void) const
-{
-	if (m_pMainForm == NULL)
-		return NULL;
-
-	return m_pMainForm->client();
-}
-
-
 // Create a new sampler channel, if not already.
-bool qsamplerChannel::addChannel (void)
+bool Channel::addChannel (void)
 {
-	if (client() == NULL)
+	MainForm* pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL)
 		return false;
 
 	// Are we a new channel?
 	if (m_iChannelID < 0) {
-		m_iChannelID = ::lscp_add_channel(client());
+		m_iChannelID = ::lscp_add_channel(pMainForm->client());
 		if (m_iChannelID < 0) {
 			appendMessagesClient("lscp_add_channel");
-			appendMessagesError(QObject::tr("Could not add channel.\n\nSorry."));
+			appendMessagesError(
+				QObject::tr("Could not add channel.\n\nSorry."));
 		}   // Otherwise it's created...
 		else appendMessages(QObject::tr("added."));
 	}
@@ -116,14 +99,17 @@ bool qsamplerChannel::addChannel (void)
 
 
 // Remove sampler channel.
-bool qsamplerChannel::removeChannel (void)
+bool Channel::removeChannel (void)
 {
-	if (client() == NULL)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL)
 		return false;
 
 	// Are we an existing channel?
 	if (m_iChannelID >= 0) {
-		if (::lscp_remove_channel(client(), m_iChannelID) != LSCP_OK) {
+		if (::lscp_remove_channel(pMainForm->client(), m_iChannelID) != LSCP_OK) {
 			appendMessagesClient("lscp_remove_channel");
 			appendMessagesError(QObject::tr("Could not remove channel.\n\nSorry."));
 		} else {
@@ -139,41 +125,46 @@ bool qsamplerChannel::removeChannel (void)
 
 
 // Channel-ID (aka Sammpler-Channel) accessors.
-int qsamplerChannel::channelID (void) const
+int Channel::channelID (void) const
 {
 	return m_iChannelID;
 }
 
-void qsamplerChannel::setChannelID ( int iChannelID )
+void Channel::setChannelID ( int iChannelID )
 {
 	m_iChannelID = iChannelID;
 }
 
 
 // Readable channel name.
-QString qsamplerChannel::channelName (void) const
+QString Channel::channelName (void) const
 {
 	return (m_iChannelID < 0 ? QObject::tr("New Channel") : QObject::tr("Channel %1").arg(m_iChannelID));
 }
 
 
 // Engine name accessors.
-const QString& qsamplerChannel::engineName (void) const
+const QString& Channel::engineName (void) const
 {
 	return m_sEngineName;
 }
 
-bool qsamplerChannel::loadEngine ( const QString& sEngineName )
+bool Channel::loadEngine ( const QString& sEngineName )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (m_iInstrumentStatus == 100 && m_sEngineName == sEngineName)
 		return true;
 
-	if (::lscp_load_engine(client(), sEngineName.latin1(), m_iChannelID) != LSCP_OK) {
+	if (::lscp_load_engine(pMainForm->client(),
+			sEngineName.toUtf8().constData(), m_iChannelID) != LSCP_OK) {
 		appendMessagesClient("lscp_load_engine");
 		return false;
 	}
+
 	appendMessages(QObject::tr("Engine: %1.").arg(sEngineName));
 
 	m_sEngineName = sEngineName;
@@ -182,40 +173,50 @@ bool qsamplerChannel::loadEngine ( const QString& sEngineName )
 
 
 // Instrument filename accessor.
-const QString& qsamplerChannel::instrumentFile (void) const
+const QString& Channel::instrumentFile (void) const
 {
 	return m_sInstrumentFile;
 }
 
 // Instrument index accessor.
-int qsamplerChannel::instrumentNr (void) const
+int Channel::instrumentNr (void) const
 {
 	return m_iInstrumentNr;
 }
 
 // Instrument name accessor.
-const QString& qsamplerChannel::instrumentName (void) const
+const QString& Channel::instrumentName (void) const
 {
 	return m_sInstrumentName;
 }
 
 // Instrument status accessor.
-int qsamplerChannel::instrumentStatus (void) const
+int Channel::instrumentStatus (void) const
 {
 	return m_iInstrumentStatus;
 }
 
 // Instrument file loader.
-bool qsamplerChannel::loadInstrument ( const QString& sInstrumentFile, int iInstrumentNr )
+bool Channel::loadInstrument ( const QString& sInstrumentFile, int iInstrumentNr )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (!isInstrumentFile(sInstrumentFile))
 		return false;
 	if (m_iInstrumentStatus == 100 && m_sInstrumentFile == sInstrumentFile && m_iInstrumentNr == iInstrumentNr)
 		return true;
 
-	if (::lscp_load_instrument_non_modal(client(), sInstrumentFile.latin1(), iInstrumentNr, m_iChannelID) != LSCP_OK) {
+	if (
+		::lscp_load_instrument_non_modal(
+			pMainForm->client(),
+			qsamplerUtilities::lscpEscapePath(
+				sInstrumentFile).toUtf8().constData(),
+			iInstrumentNr, m_iChannelID
+		) != LSCP_OK
+	) {
 		appendMessagesClient("lscp_load_instrument");
 		return false;
 	}
@@ -228,7 +229,7 @@ bool qsamplerChannel::loadInstrument ( const QString& sInstrumentFile, int iInst
 
 
 // Special instrument file/name/number settler.
-bool qsamplerChannel::setInstrument ( const QString& sInstrumentFile, int iInstrumentNr )
+bool Channel::setInstrument ( const QString& sInstrumentFile, int iInstrumentNr )
 {
 	m_sInstrumentFile = sInstrumentFile;
 	m_iInstrumentNr = iInstrumentNr;
@@ -244,19 +245,23 @@ bool qsamplerChannel::setInstrument ( const QString& sInstrumentFile, int iInstr
 
 
 // MIDI driver type accessors (DEPRECATED).
-const QString& qsamplerChannel::midiDriver (void) const
+const QString& Channel::midiDriver (void) const
 {
 	return m_sMidiDriver;
 }
 
-bool qsamplerChannel::setMidiDriver ( const QString& sMidiDriver )
+bool Channel::setMidiDriver ( const QString& sMidiDriver )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (m_iInstrumentStatus == 100 && m_sMidiDriver == sMidiDriver)
 		return true;
 
-	if (::lscp_set_channel_midi_type(client(), m_iChannelID, sMidiDriver.latin1()) != LSCP_OK) {
+	if (::lscp_set_channel_midi_type(pMainForm->client(),
+			m_iChannelID, sMidiDriver.toUtf8().constData()) != LSCP_OK) {
 		appendMessagesClient("lscp_set_channel_midi_type");
 		return false;
 	}
@@ -269,19 +274,22 @@ bool qsamplerChannel::setMidiDriver ( const QString& sMidiDriver )
 
 
 // MIDI device accessors.
-int qsamplerChannel::midiDevice (void) const
+int Channel::midiDevice (void) const
 {
 	return m_iMidiDevice;
 }
 
-bool qsamplerChannel::setMidiDevice ( int iMidiDevice )
+bool Channel::setMidiDevice ( int iMidiDevice )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (m_iInstrumentStatus == 100 && m_iMidiDevice == iMidiDevice)
 		return true;
 
-	if (::lscp_set_channel_midi_device(client(), m_iChannelID, iMidiDevice) != LSCP_OK) {
+	if (::lscp_set_channel_midi_device(pMainForm->client(), m_iChannelID, iMidiDevice) != LSCP_OK) {
 		appendMessagesClient("lscp_set_channel_midi_device");
 		return false;
 	}
@@ -294,19 +302,22 @@ bool qsamplerChannel::setMidiDevice ( int iMidiDevice )
 
 
 // MIDI port number accessor.
-int qsamplerChannel::midiPort (void) const
+int Channel::midiPort (void) const
 {
 	return m_iMidiPort;
 }
 
-bool qsamplerChannel::setMidiPort ( int iMidiPort )
+bool Channel::setMidiPort ( int iMidiPort )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (m_iInstrumentStatus == 100 && m_iMidiPort == iMidiPort)
 		return true;
 
-	if (::lscp_set_channel_midi_port(client(), m_iChannelID, iMidiPort) != LSCP_OK) {
+	if (::lscp_set_channel_midi_port(pMainForm->client(), m_iChannelID, iMidiPort) != LSCP_OK) {
 		appendMessagesClient("lscp_set_channel_midi_port");
 		return false;
 	}
@@ -319,19 +330,22 @@ bool qsamplerChannel::setMidiPort ( int iMidiPort )
 
 
 // MIDI channel accessor.
-int qsamplerChannel::midiChannel (void) const
+int Channel::midiChannel (void) const
 {
 	return m_iMidiChannel;
 }
 
-bool qsamplerChannel::setMidiChannel ( int iMidiChannel )
+bool Channel::setMidiChannel ( int iMidiChannel )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (m_iInstrumentStatus == 100 && m_iMidiChannel == iMidiChannel)
 		return true;
 
-	if (::lscp_set_channel_midi_channel(client(), m_iChannelID, iMidiChannel) != LSCP_OK) {
+	if (::lscp_set_channel_midi_channel(pMainForm->client(), m_iChannelID, iMidiChannel) != LSCP_OK) {
 		appendMessagesClient("lscp_set_channel_midi_channel");
 		return false;
 	}
@@ -343,20 +357,51 @@ bool qsamplerChannel::setMidiChannel ( int iMidiChannel )
 }
 
 
+// MIDI instrument map accessor.
+int Channel::midiMap (void) const
+{
+	return m_iMidiMap;
+}
+
+bool Channel::setMidiMap ( int iMidiMap )
+{
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
+		return false;
+	if (m_iInstrumentStatus == 100 && m_iMidiMap == iMidiMap)
+		return true;
+#ifdef CONFIG_MIDI_INSTRUMENT
+	if (::lscp_set_channel_midi_map(pMainForm->client(), m_iChannelID, iMidiMap) != LSCP_OK) {
+		appendMessagesClient("lscp_set_channel_midi_map");
+		return false;
+	}
+#endif
+	appendMessages(QObject::tr("MIDI map: %1.").arg(iMidiMap));
+
+	m_iMidiMap = iMidiMap;
+	return true;
+}
+
+
 // Audio device accessor.
-int qsamplerChannel::audioDevice (void) const
+int Channel::audioDevice (void) const
 {
 	return m_iAudioDevice;
 }
 
-bool qsamplerChannel::setAudioDevice ( int iAudioDevice )
+bool Channel::setAudioDevice ( int iAudioDevice )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (m_iInstrumentStatus == 100 && m_iAudioDevice == iAudioDevice)
 		return true;
 
-	if (::lscp_set_channel_audio_device(client(), m_iChannelID, iAudioDevice) != LSCP_OK) {
+	if (::lscp_set_channel_audio_device(pMainForm->client(), m_iChannelID, iAudioDevice) != LSCP_OK) {
 		appendMessagesClient("lscp_set_channel_audio_device");
 		return false;
 	}
@@ -369,19 +414,23 @@ bool qsamplerChannel::setAudioDevice ( int iAudioDevice )
 
 
 // Audio driver type accessors (DEPRECATED).
-const QString& qsamplerChannel::audioDriver (void) const
+const QString& Channel::audioDriver (void) const
 {
 	return m_sAudioDriver;
 }
 
-bool qsamplerChannel::setAudioDriver ( const QString& sAudioDriver )
+bool Channel::setAudioDriver ( const QString& sAudioDriver )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (m_iInstrumentStatus == 100 && m_sAudioDriver == sAudioDriver)
 		return true;
 
-	if (::lscp_set_channel_audio_type(client(), m_iChannelID, sAudioDriver.latin1()) != LSCP_OK) {
+	if (::lscp_set_channel_audio_type(pMainForm->client(),
+			m_iChannelID, sAudioDriver.toUtf8().constData()) != LSCP_OK) {
 		appendMessagesClient("lscp_set_channel_audio_type");
 		return false;
 	}
@@ -394,19 +443,22 @@ bool qsamplerChannel::setAudioDriver ( const QString& sAudioDriver )
 
 
 // Channel volume accessors.
-float qsamplerChannel::volume (void) const
+float Channel::volume (void) const
 {
 	return m_fVolume;
 }
 
-bool qsamplerChannel::setVolume ( float fVolume )
+bool Channel::setVolume ( float fVolume )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (m_iInstrumentStatus == 100 && m_fVolume == fVolume)
 		return true;
 
-	if (::lscp_set_channel_volume(client(), m_iChannelID, fVolume) != LSCP_OK) {
+	if (::lscp_set_channel_volume(pMainForm->client(), m_iChannelID, fVolume) != LSCP_OK) {
 		appendMessagesClient("lscp_set_channel_volume");
 		return false;
 	}
@@ -419,20 +471,23 @@ bool qsamplerChannel::setVolume ( float fVolume )
 
 
 // Sampler channel mute state.
-bool qsamplerChannel::channelMute (void) const
+bool Channel::channelMute (void) const
 {
 	return m_bMute;
 }
 
-bool qsamplerChannel::setChannelMute ( bool bMute )
+bool Channel::setChannelMute ( bool bMute )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (m_iInstrumentStatus == 100 && ((m_bMute && bMute) || (!m_bMute && !bMute)))
 		return true;
 
 #ifdef CONFIG_MUTE_SOLO
-	if (::lscp_set_channel_mute(client(), m_iChannelID, bMute) != LSCP_OK) {
+	if (::lscp_set_channel_mute(pMainForm->client(), m_iChannelID, bMute) != LSCP_OK) {
 		appendMessagesClient("lscp_set_channel_mute");
 		return false;
 	}
@@ -446,20 +501,23 @@ bool qsamplerChannel::setChannelMute ( bool bMute )
 
 
 // Sampler channel solo state.
-bool qsamplerChannel::channelSolo (void) const
+bool Channel::channelSolo (void) const
 {
 	return m_bSolo;
 }
 
-bool qsamplerChannel::setChannelSolo ( bool bSolo )
+bool Channel::setChannelSolo ( bool bSolo )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (m_iInstrumentStatus == 100 && ((m_bSolo && bSolo) || (!m_bSolo && !bSolo)))
 		return true;
 
 #ifdef CONFIG_MUTE_SOLO
-	if (::lscp_set_channel_solo(client(), m_iChannelID, bSolo) != LSCP_OK) {
+	if (::lscp_set_channel_solo(pMainForm->client(), m_iChannelID, bSolo) != LSCP_OK) {
 		appendMessagesClient("lscp_set_channel_solo");
 		return false;
 	}
@@ -473,20 +531,23 @@ bool qsamplerChannel::setChannelSolo ( bool bSolo )
 
 
 // Audio routing accessors.
-int qsamplerChannel::audioChannel ( int iAudioOut ) const
+int Channel::audioChannel ( int iAudioOut ) const
 {
 	return m_audioRouting[iAudioOut];
 }
 
-bool qsamplerChannel::setAudioChannel ( int iAudioOut, int iAudioIn )
+bool Channel::setAudioChannel ( int iAudioOut, int iAudioIn )
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 	if (m_iInstrumentStatus == 100 &&
 			m_audioRouting[iAudioOut] == iAudioIn)
 		return true;
 
-	if (::lscp_set_channel_audio_channel(client(),
+	if (::lscp_set_channel_audio_channel(pMainForm->client(),
 			m_iChannelID, iAudioOut, iAudioIn) != LSCP_OK) {
 		appendMessagesClient("lscp_set_channel_audio_channel");
 		return false;
@@ -500,14 +561,14 @@ bool qsamplerChannel::setAudioChannel ( int iAudioOut, int iAudioIn )
 }
 
 // The audio routing map itself.
-const qsamplerChannelRoutingMap& qsamplerChannel::audioRouting (void) const
+const ChannelRoutingMap& Channel::audioRouting (void) const
 {
 	return m_audioRouting;
 }
 
 
 // Istrument name remapper.
-void qsamplerChannel::updateInstrumentName (void)
+void Channel::updateInstrumentName (void)
 {
 #ifndef CONFIG_INSTRUMENT_NAME
 	m_sInstrumentName = getInstrumentName(m_sInstrumentFile,
@@ -517,13 +578,16 @@ void qsamplerChannel::updateInstrumentName (void)
 
 
 // Update whole channel info state.
-bool qsamplerChannel::updateChannelInfo (void)
+bool Channel::updateChannelInfo (void)
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 
 	// Read channel information.
-	lscp_channel_info_t *pChannelInfo = ::lscp_get_channel_info(client(), m_iChannelID);
+	lscp_channel_info_t *pChannelInfo = ::lscp_get_channel_info(pMainForm->client(), m_iChannelID);
 	if (pChannelInfo == NULL) {
 		appendMessagesClient("lscp_get_channel_info");
 		appendMessagesError(QObject::tr("Could not get channel information.\n\nSorry."));
@@ -532,9 +596,11 @@ bool qsamplerChannel::updateChannelInfo (void)
 
 #ifdef CONFIG_INSTRUMENT_NAME
 	// We got all actual instrument datum...
-	m_sInstrumentFile = pChannelInfo->instrument_file;
+	m_sInstrumentFile =
+		qsamplerUtilities::lscpEscapedPathToPosix(pChannelInfo->instrument_file);
 	m_iInstrumentNr   = pChannelInfo->instrument_nr;
-	m_sInstrumentName = pChannelInfo->instrument_name;
+	m_sInstrumentName =
+		qsamplerUtilities::lscpEscapedTextToRaw(pChannelInfo->instrument_name);
 #else
 	// First, check if intrument name has changed,
 	// taking care that instrument name lookup might be expensive,
@@ -552,6 +618,9 @@ bool qsamplerChannel::updateChannelInfo (void)
 	m_iMidiDevice       = pChannelInfo->midi_device;
 	m_iMidiPort         = pChannelInfo->midi_port;
 	m_iMidiChannel      = pChannelInfo->midi_channel;
+#ifdef CONFIG_MIDI_INSTRUMENT
+	m_iMidiMap          = pChannelInfo->midi_map;
+#endif
 	m_iAudioDevice      = pChannelInfo->audio_device;
 	m_fVolume           = pChannelInfo->volume;
 #ifdef CONFIG_MUTE_SOLO
@@ -570,7 +639,7 @@ bool qsamplerChannel::updateChannelInfo (void)
 	lscp_device_info_t *pDeviceInfo;
 	const QString sNone = QObject::tr("(none)");
 	// Audio device driver type.
-	pDeviceInfo = ::lscp_get_audio_device_info(client(), m_iAudioDevice);
+	pDeviceInfo = ::lscp_get_audio_device_info(pMainForm->client(), m_iAudioDevice);
 	if (pDeviceInfo == NULL) {
 		appendMessagesClient("lscp_get_audio_device_info");
 		m_sAudioDriver = sNone;
@@ -578,7 +647,7 @@ bool qsamplerChannel::updateChannelInfo (void)
 		m_sAudioDriver = pDeviceInfo->driver;
 	}
 	// MIDI device driver type.
-	pDeviceInfo = ::lscp_get_midi_device_info(client(), m_iMidiDevice);
+	pDeviceInfo = ::lscp_get_midi_device_info(pMainForm->client(), m_iMidiDevice);
 	if (pDeviceInfo == NULL) {
 		appendMessagesClient("lscp_get_midi_device_info");
 		m_sMidiDriver = sNone;
@@ -588,22 +657,30 @@ bool qsamplerChannel::updateChannelInfo (void)
 
 	// Set the audio routing map.
 	m_audioRouting.clear();
-	char **ppszRouting = pChannelInfo->audio_routing;
-	for (int i = 0; ppszRouting && ppszRouting[i]; i++) {
-		m_audioRouting[i] = ::atoi(ppszRouting[i]);
-	}
+#ifdef CONFIG_AUDIO_ROUTING
+	int *piAudioRouting = pChannelInfo->audio_routing;
+	for (int i = 0; piAudioRouting && piAudioRouting[i] >= 0; i++)
+		m_audioRouting[i] = piAudioRouting[i];
+#else
+	char **ppszAudioRouting = pChannelInfo->audio_routing;
+	for (int i = 0; ppszAudioRouting && ppszAudioRouting[i]; i++)
+		m_audioRouting[i] = ::atoi(ppszAudioRouting[i]);
+#endif
 
 	return true;
 }
 
 
 // Reset channel method.
-bool qsamplerChannel::channelReset (void)
+bool Channel::channelReset (void)
 {
-	if (client() == NULL || m_iChannelID < 0)
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
 		return false;
 
-	if (::lscp_reset_channel(client(), m_iChannelID) != LSCP_OK) {
+	if (::lscp_reset_channel(pMainForm->client(), m_iChannelID) != LSCP_OK) {
 		appendMessagesClient("lscp_reset_channel");
 		return false;
 	}
@@ -614,14 +691,61 @@ bool qsamplerChannel::channelReset (void)
 }
 
 
-// Channel setup dialog form.
-bool qsamplerChannel::channelSetup ( QWidget *pParent )
+// Spawn instrument editor method.
+bool Channel::editChannel (void)
 {
+#ifdef CONFIG_EDIT_INSTRUMENT
+
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+	if (pMainForm->client() == NULL || m_iChannelID < 0)
+		return false;
+
+	if (::lscp_edit_channel_instrument(pMainForm->client(), m_iChannelID)
+		!= LSCP_OK) {
+		appendMessagesClient("lscp_edit_channel_instrument");
+		appendMessagesError(QObject::tr(
+			"Could not launch an appropriate instrument editor "
+			"for the given instrument!\n\n"
+			"Make sure you have an appropriate "
+			"instrument editor like 'gigedit' installed "
+			"and that it placed its mandatory DLL file "
+			"into the sampler's plugin directory.")
+		);
+		return false;
+	}
+
+	appendMessages(QObject::tr("edit instrument."));
+
+	return true;
+
+#else
+
+	appendMessagesError(QObject::tr(
+		"Sorry, QSampler was compiled for a version of liblscp "
+		"which lacks this feature.\n\n"
+		"You may want to update liblscp and recompile QSampler afterwards.")
+	);
+
+	return false;
+
+#endif
+}
+
+
+// Channel setup dialog form.
+bool Channel::channelSetup ( QWidget *pParent )
+{
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm == NULL)
+		return false;
+
 	bool bResult = false;
 
 	appendMessages(QObject::tr("setup..."));
 
-	qsamplerChannelForm *pChannelForm = new qsamplerChannelForm(pParent);
+	ChannelForm *pChannelForm = new ChannelForm(pParent);
 	if (pChannelForm) {
 		pChannelForm->setup(this);
 		bResult = pChannelForm->exec();
@@ -633,55 +757,61 @@ bool qsamplerChannel::channelSetup ( QWidget *pParent )
 
 
 // Redirected messages output methods.
-void qsamplerChannel::appendMessages( const QString& s ) const
+void Channel::appendMessages( const QString& s ) const
 {
-	if (m_pMainForm)
-		m_pMainForm->appendMessages(channelName() + ' ' + s);
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm)
+		pMainForm->appendMessages(channelName() + ' ' + s);
 }
 
-void qsamplerChannel::appendMessagesColor( const QString& s,
+void Channel::appendMessagesColor( const QString& s,
 	const QString& c ) const
 {
-	if (m_pMainForm)
-		m_pMainForm->appendMessagesColor(channelName() + ' ' + s, c);
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm)
+		pMainForm->appendMessagesColor(channelName() + ' ' + s, c);
 }
 
-void qsamplerChannel::appendMessagesText( const QString& s ) const
+void Channel::appendMessagesText( const QString& s ) const
 {
-	if (m_pMainForm)
-		m_pMainForm->appendMessagesText(channelName() + ' ' + s);
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm)
+		pMainForm->appendMessagesText(channelName() + ' ' + s);
 }
 
-void qsamplerChannel::appendMessagesError( const QString& s ) const
+void Channel::appendMessagesError( const QString& s ) const
 {
-	if (m_pMainForm)
-		m_pMainForm->appendMessagesError(channelName() + "\n\n" + s);
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm)
+		pMainForm->appendMessagesError(channelName() + "\n\n" + s);
 }
 
-void qsamplerChannel::appendMessagesClient( const QString& s ) const
+void Channel::appendMessagesClient( const QString& s ) const
 {
-	if (m_pMainForm)
-		m_pMainForm->appendMessagesClient(channelName() + ' ' + s);
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm)
+		pMainForm->appendMessagesClient(channelName() + ' ' + s);
 }
 
 
 // Context menu event handler.
-void qsamplerChannel::contextMenuEvent( QContextMenuEvent *pEvent )
+void Channel::contextMenuEvent( QContextMenuEvent *pEvent )
 {
-	if (m_pMainForm)
-		m_pMainForm->contextMenuEvent(pEvent);
+	MainForm *pMainForm = MainForm::getInstance();
+	if (pMainForm)
+		pMainForm->contextMenuEvent(pEvent);
 }
 
 
 // FIXME: Check whether a given file is an instrument file.
-bool qsamplerChannel::isInstrumentFile ( const QString& sInstrumentFile )
+bool Channel::isInstrumentFile ( const QString& sInstrumentFile )
 {
 	bool bResult = false;
 
 	QFile file(sInstrumentFile);
-	if (file.open(IO_ReadOnly)) {
+	if (file.open(QIODevice::ReadOnly)) {
 		char achHeader[16];
-		if (file.readBlock(achHeader, 16)) {
+		if (file.read(achHeader, 16) > 0) {
 			bResult = (::memcmp(&achHeader[0], "RIFF", 4)     == 0
 					&& ::memcmp(&achHeader[8], "DLS LIST", 8) == 0);
 		}
@@ -693,7 +823,7 @@ bool qsamplerChannel::isInstrumentFile ( const QString& sInstrumentFile )
 
 
 // Retrieve the instrument list of a instrument file (.gig).
-QStringList qsamplerChannel::getInstrumentList( const QString& sInstrumentFile,
+QStringList Channel::getInstrumentList( const QString& sInstrumentFile,
 	bool bInstrumentNames )
 {
 	QString sInstrumentName = QFileInfo(sInstrumentFile).fileName();
@@ -702,8 +832,13 @@ QStringList qsamplerChannel::getInstrumentList( const QString& sInstrumentFile,
 	if (isInstrumentFile(sInstrumentFile)) {
 #ifdef CONFIG_LIBGIG
 		if (bInstrumentNames) {
-			RIFF::File *pRiff = new RIFF::File(sInstrumentFile.latin1());
+			RIFF::File *pRiff
+				= new RIFF::File(sInstrumentFile.toUtf8().constData());
 			gig::File  *pGig  = new gig::File(pRiff);
+#if HAVE_LIBGIG_SETAUTOLOAD
+			// prevent sleepy response time on large .gig files
+			pGig->SetAutoLoad(false);
+#endif
 			gig::Instrument *pInstrument = pGig->GetFirstInstrument();
 			while (pInstrument) {
 				instlist.append((pInstrument->pInfo)->Name.c_str());
@@ -724,7 +859,7 @@ QStringList qsamplerChannel::getInstrumentList( const QString& sInstrumentFile,
 
 
 // Retrieve the spacific instrument name of a instrument file (.gig), given its index.
-QString qsamplerChannel::getInstrumentName( const QString& sInstrumentFile,
+QString Channel::getInstrumentName( const QString& sInstrumentFile,
 	int iInstrumentNr, bool bInstrumentNames )
 {
 	QString sInstrumentName;
@@ -733,8 +868,13 @@ QString qsamplerChannel::getInstrumentName( const QString& sInstrumentFile,
 		sInstrumentName = QFileInfo(sInstrumentFile).fileName();
 #ifdef CONFIG_LIBGIG
 		if (bInstrumentNames) {
-			RIFF::File *pRiff = new RIFF::File(sInstrumentFile.latin1());
-			gig::File  *pGig  = new gig::File(pRiff);
+			RIFF::File *pRiff
+				= new RIFF::File(sInstrumentFile.toUtf8().constData());
+			gig::File *pGig = new gig::File(pRiff);
+#if HAVE_LIBGIG_SETAUTOLOAD
+			// prevent sleepy response time on large .gig files
+			pGig->SetAutoLoad(false);
+#endif
 			int iIndex = 0;
 			gig::Instrument *pInstrument = pGig->GetFirstInstrument();
 			while (pInstrument) {
@@ -759,172 +899,174 @@ QString qsamplerChannel::getInstrumentName( const QString& sInstrumentFile,
 
 
 // Common invalid name-helpers.
-QString qsamplerChannel::noEngineName (void)
+QString Channel::noEngineName (void)
 {
 	return QObject::tr("(No engine)");
 }
 
-QString qsamplerChannel::noInstrumentName (void)
+QString Channel::noInstrumentName (void)
 {
 	return QObject::tr("(No instrument)");
 }
 
-QString qsamplerChannel::loadingInstrument (void) {
+QString Channel::loadingInstrument (void) {
 	return QObject::tr("(Loading instrument...)");
 }
 
 
-
 //-------------------------------------------------------------------------
-// qsamplerChannelRoutingTable - Channel routing table.
-//
+// QSampler::ChannelRoutingModel - data model for audio routing
+//                                 (used for QTableView)
 
-// Constructor.
-qsamplerChannelRoutingTable::qsamplerChannelRoutingTable (
-	QWidget *pParent, const char *pszName )
-	: QTable(pParent, pszName)
-{
-	// Set fixed number of columns.
-	QTable::setNumCols(2);
-	QTable::setShowGrid(false);
-	QTable::setSorting(false);
-	QTable::setFocusStyle(QTable::FollowStyle);
-	QTable::setSelectionMode(QTable::NoSelection);
-	// No vertical header.
-	QTable::verticalHeader()->hide();
-	QTable::setLeftMargin(0);
-	// Initialize the fixed table column headings.
-	QHeader *pHeader = QTable::horizontalHeader();
-	pHeader->setLabel(0, tr("Sampler Channel"));
-	pHeader->setLabel(1, tr("Device Channel"));
-	// Set read-onlyness of each column
-	QTable::setColumnReadOnly(0, true);
-//	QTable::setColumnReadOnly(1, false); -- of course not.
-	QTable::setColumnStretchable(1, true);
-}
-
-// Default destructor.
-qsamplerChannelRoutingTable::~qsamplerChannelRoutingTable (void)
+ChannelRoutingModel::ChannelRoutingModel ( QObject *pParent )
+	: QAbstractTableModel(pParent), m_pDevice(NULL)
 {
 }
 
 
-// Routing map table renderer.
-void qsamplerChannelRoutingTable::refresh ( qsamplerDevice *pDevice,
-	const qsamplerChannelRoutingMap& routing )
+int ChannelRoutingModel::rowCount ( const QModelIndex& /*parent*/) const
 {
-	if (pDevice == NULL)
-		return;
+	return m_routing.size();
+}
 
-	// Always (re)start it empty.
-	QTable::setUpdatesEnabled(false);
-	QTable::setNumRows(0);
+
+int ChannelRoutingModel::columnCount ( const QModelIndex& /*parent*/) const
+{
+	return 1;
+}
+
+
+Qt::ItemFlags ChannelRoutingModel::flags ( const QModelIndex& /*index*/) const
+{
+	return Qt::ItemIsSelectable | Qt::ItemIsEditable | Qt::ItemIsEnabled;
+}
+
+
+bool ChannelRoutingModel::setData ( const QModelIndex& index,
+	const QVariant& value, int /*role*/)
+{
+	if (!index.isValid())
+		return false;
+
+	m_routing[index.row()] = value.toInt();
+
+	emit dataChanged(index, index);
+	return true;
+}
+
+
+QVariant ChannelRoutingModel::data ( const QModelIndex &index, int role ) const
+{
+	if (!index.isValid())
+		return QVariant();
+	if (role != Qt::DisplayRole)
+		return QVariant();
+	if (index.column() != 0)
+		return QVariant();
+
+	ChannelRoutingItem item;
 
 	// The common device port item list.
-	QStringList opts;
-	qsamplerDevicePortList& ports = pDevice->ports();
-	qsamplerDevicePort *pPort;
-	for (pPort = ports.first(); pPort; pPort = ports.next()) {
-		opts.append(pDevice->deviceTypeName()
-			+ ' ' + pDevice->driverName()
-			+ ' ' + pPort->portName());
+	DevicePortList& ports = m_pDevice->ports();
+	QListIterator<DevicePort *> iter(ports);
+	while (iter.hasNext()) {
+		DevicePort *pPort = iter.next();
+		item.options.append(
+			m_pDevice->deviceTypeName()
+			+ ' ' + m_pDevice->driverName()
+			+ ' ' + pPort->portName()
+		);
 	}
 
-	// Those items shall have a proper pixmap...
-	QPixmap pmChannel = QPixmap::fromMimeSource("qsamplerChannel.png");
-	QPixmap pmDevice;
-	switch (pDevice->deviceType()) {
-	case qsamplerDevice::Audio:
-		pmDevice = QPixmap::fromMimeSource("audio2.png");
-		break;
-	case qsamplerDevice::Midi:
-		pmDevice = QPixmap::fromMimeSource("midi2.png");
-		break;
-	case qsamplerDevice::None:
-		break;
-	}
+	item.selection = m_routing[index.row()];
 
-	// Fill the routing table...
-	QTable::insertRows(0, routing.count());
-	int iRow = 0;
-	qsamplerChannelRoutingMap::ConstIterator iter;
-	for (iter = routing.begin(); iter != routing.end(); ++iter) {
-		QTable::setPixmap(iRow, 0, pmChannel);
-		QTable::setText(iRow, 0, pDevice->deviceTypeName()
-			+ ' ' + QString::number(iter.key()));
-		qsamplerChannelRoutingComboBox *pComboItem =
-			new qsamplerChannelRoutingComboBox(this, opts, pmDevice);
-		pComboItem->setCurrentItem(iter.data());
-		QTable::setItem(iRow, 1, pComboItem);
-		++iRow;
-	}
-
-	// Adjust optimal column widths.
-	QTable::adjustColumn(0);
-	QTable::adjustColumn(1);
-
-	QTable::setUpdatesEnabled(true);
-	QTable::updateContents();
+	return QVariant::fromValue(item);
 }
 
 
-// Commit any pending editing.
-void qsamplerChannelRoutingTable::flush (void)
+QVariant ChannelRoutingModel::headerData ( int section,
+	Qt::Orientation orientation, int role) const
 {
-	if (QTable::isEditing())
-	    QTable::endEdit(QTable::currEditRow(), QTable::currEditCol(), true, true);
+	if (role != Qt::DisplayRole)
+		return QVariant();
+
+	switch (orientation) {
+		case Qt::Horizontal:
+			return UNICODE_RIGHT_ARROW + QObject::tr(" Device Channel");
+		case Qt::Vertical:
+			return QObject::tr("Sampler Channel ") +
+				QString::number(section) + " " + UNICODE_RIGHT_ARROW;
+		default:
+			return QVariant();
+	}
+}
+
+
+void ChannelRoutingModel::refresh ( Device *pDevice,
+	const ChannelRoutingMap& routing )
+{
+	m_pDevice = pDevice;
+	m_routing = routing;
+	// inform the outer world (QTableView) that our data changed
+	QAbstractTableModel::reset();
 }
 
 
 //-------------------------------------------------------------------------
-// qsamplerChannelRoutingComboBox - Custom combo box for routing table.
+// QSampler::ChannelRoutingDelegate - table cell renderer for audio routing
 //
 
-// Constructor.
-qsamplerChannelRoutingComboBox::qsamplerChannelRoutingComboBox (
-	QTable *pTable, const QStringList& list, const QPixmap& pixmap )
-	: QTableItem(pTable, QTableItem::WhenCurrent, QString::null, pixmap),
-	m_list(list)
+ChannelRoutingDelegate::ChannelRoutingDelegate ( QObject *pParent )
+	: QItemDelegate(pParent)
 {
-	m_iCurrentItem = 0;
 }
 
-// Public accessors.
-void qsamplerChannelRoutingComboBox::setCurrentItem ( int iCurrentItem )
-{
-	m_iCurrentItem = iCurrentItem;
 
-	QTableItem::setText(m_list[iCurrentItem]);
-}
-
-int qsamplerChannelRoutingComboBox::currentItem (void) const
+QWidget* ChannelRoutingDelegate::createEditor ( QWidget *pParent,
+	const QStyleOptionViewItem & option, const QModelIndex& index ) const
 {
-	return m_iCurrentItem;
-}
+	if (!index.isValid())
+		return NULL;
 
-// Virtual implemetations.
-QWidget *qsamplerChannelRoutingComboBox::createEditor (void) const
-{
-	QComboBox *pComboBox = new QComboBox(QTableItem::table()->viewport());
-	QObject::connect(pComboBox, SIGNAL(activated(int)),
-		QTableItem::table(), SLOT(doValueChanged()));
-	for (QStringList::ConstIterator iter = m_list.begin();
-			iter != m_list.end(); iter++) {
-		pComboBox->insertItem(QTableItem::pixmap(), *iter);
-	}
-	pComboBox->setCurrentItem(m_iCurrentItem);
+	if (index.column() != 0)
+		return NULL;
+
+	ChannelRoutingItem item = index.model()->data(index, Qt::DisplayRole).value<ChannelRoutingItem>();
+
+	QComboBox* pComboBox = new QComboBox(pParent);
+	pComboBox->addItems(item.options);
+	pComboBox->setCurrentIndex(item.selection);
+	pComboBox->setEnabled(true);
+	pComboBox->setGeometry(option.rect);
 	return pComboBox;
 }
 
-void qsamplerChannelRoutingComboBox::setContentFromEditor ( QWidget *pWidget )
+
+void ChannelRoutingDelegate::setEditorData ( QWidget *pEditor,
+	const QModelIndex &index) const
 {
-	if (pWidget->inherits("QComboBox")) {
-		QComboBox *pComboBox = (QComboBox *) pWidget;
-		m_iCurrentItem = pComboBox->currentItem();
-		QTableItem::setText(pComboBox->currentText());
-	}
-	else QTableItem::setContentFromEditor(pWidget);
+	ChannelRoutingItem item = index.model()->data(index,
+		Qt::DisplayRole).value<ChannelRoutingItem> ();
+	QComboBox* pComboBox = static_cast<QComboBox*> (pEditor);
+	pComboBox->setCurrentIndex(item.selection);
 }
 
 
+void ChannelRoutingDelegate::setModelData ( QWidget* pEditor,
+	QAbstractItemModel *pModel, const QModelIndex& index ) const
+{
+	QComboBox *pComboBox = static_cast<QComboBox*> (pEditor);
+	pModel->setData(index, pComboBox->currentIndex());
+}
+
+
+void ChannelRoutingDelegate::updateEditorGeometry ( QWidget *pEditor,
+	const QStyleOptionViewItem& option, const QModelIndex &/* index */) const
+{
+	pEditor->setGeometry(option.rect);
+}
+
+} // namespace QSampler
+
+
 // end of qsamplerChannel.cpp
